@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2018 Artifex Software, Inc.
+/* Copyright (C) 2001-2019 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -33,12 +33,6 @@
 #include "gdevp14.h"        /* Needed to patch up the procs after compositor creation */
 #include "gdevsclass.h"
 
-/* This seems to be broadly similar to the 'device filter stack' which is defined in gsdfilt.c
- * and the stack for which is defined in the graphics state (dfilter_stack) but which seems to be
- * completely unused. We should probably remove dfilter_stack from the graphics state and remove
- * gsdfilt.c from the build.
- */
-
 /*
  * It would be nice if we could rewrite the clist handling to use this kind of device class chain
  * instead of the nasty hackery it currently utilises (stores the device procs for the existing
@@ -99,13 +93,21 @@
  */
 int default_subclass_open_device(gx_device *dev)
 {
-    if (dev->child) {
-        dev_proc(dev->child, open_device)(dev->child);
-        dev->child->is_open = true;
-        gx_update_from_subclass(dev);
-    }
+    int code = 0;
 
-    return 0;
+    /* observed with Bug 699794, don't set is_open = true if the open_device failed */
+    /* and make sure to propagate the return code from the child device to caller.  */
+    /* Only open the child if it was closed  and if child open is OK, return 1.     */
+    /* (see gs_opendevice) */
+    if (dev->child && dev->child->is_open == 0) {
+        code = dev_proc(dev->child, open_device)(dev->child);
+        if (code >= 0) {
+            dev->child->is_open = true;
+            code = 1;	/* device had been closed, but now is open */
+        }
+        gx_update_from_subclass(dev);	/* this is probably safe to do even if the open failed */
+    }
+    return code;
 }
 
 void default_subclass_get_initial_matrix(gx_device *dev, gs_matrix *pmat)
@@ -538,6 +540,7 @@ int default_subclass_create_compositor(gx_device *dev, gx_device **pcdev, const
                     dev->color_info = dev->child->color_info;
 
                     psubclass_data->saved_compositor_method = p14dev->procs.create_compositor;
+                    psubclass_data->forwarding_dev = fdev;
                     p14dev->procs.create_compositor = gx_subclass_create_compositor;
                 }
 
@@ -783,13 +786,18 @@ int default_subclass_pop_transparency_state(gx_device *dev, gs_gstate *pgs)
     return 0;
 }
 
-int default_subclass_put_image(gx_device *dev, const byte **buffers, int num_chan, int x, int y,
+int default_subclass_put_image(gx_device *dev, gx_device *mdev, const byte **buffers, int num_chan, int x, int y,
             int width, int height, int row_stride,
             int alpha_plane_index, int tag_plane_index)
 {
-    if (dev->child)
-        return dev_proc(dev->child, put_image)(dev->child, buffers, num_chan, x, y, width, height, row_stride, alpha_plane_index, tag_plane_index);
-
+    if (dev->child) {
+        if (dev == mdev) {
+            return dev_proc(dev->child, put_image)(dev->child, dev->child, buffers, num_chan, x, y, width, height, row_stride, alpha_plane_index, tag_plane_index);
+        }
+        else {
+            return dev_proc(dev->child, put_image)(dev->child, mdev, buffers, num_chan, x, y, width, height, row_stride, alpha_plane_index, tag_plane_index);
+        }
+    }
     return 0;
 }
 
@@ -885,3 +893,53 @@ int default_subclass_process_page(gx_device *dev, gx_process_page_options_t *opt
 
     return 0;
 }
+
+int default_subclass_fill_stroke_path(gx_device *dev, const gs_gstate *pgs, gx_path *ppath,
+        const gx_fill_params *fill_params, const gx_drawing_color *pdcolor_fill,
+        const gx_stroke_params *stroke_params, const gx_drawing_color *pdcolor_stroke,
+        const gx_clip_path *pcpath)
+{
+    if (dev->child)
+        return dev_proc(dev->child, fill_stroke_path)(dev, pgs, ppath, fill_params, pdcolor_fill,
+                                                      stroke_params, pdcolor_stroke, pcpath);
+    return 0;
+}
+
+int default_subclass_transform_pixel_region(gx_device *dev, transform_pixel_region_reason reason, transform_pixel_region_data *data)
+{
+    if (dev->child)
+        return dev_proc(dev->child, transform_pixel_region)(dev->child, reason, data);
+
+    return gs_error_unknownerror;
+}
+
+void default_subclass_finalize(const gs_memory_t *cmem, void *vptr)
+{
+    gx_device * const dev = (gx_device *)vptr;
+    generic_subclass_data *psubclass_data = (generic_subclass_data *)dev->subclass_data;
+    (void)cmem; /* unused */
+
+    discard(gs_closedevice(dev));
+
+    if (dev->finalize)
+        dev->finalize(dev);
+
+    if (psubclass_data) {
+        gs_free_object(dev->memory->non_gc_memory, psubclass_data, "gx_epo_finalize(suclass data)");
+        dev->subclass_data = NULL;
+    }
+    if (dev->child) {
+        gs_free_object(dev->memory->stable_memory, dev->child, "free child device memory for subclassing device");
+    }
+    if (dev->stype_is_dynamic)
+        gs_free_const_object(dev->memory->non_gc_memory, dev->stype,
+                             "default_subclass_finalize");
+    if (dev->parent)
+        dev->parent->child = dev->child;
+    if (dev->child)
+        dev->child->parent = dev->parent;
+    if (dev->icc_struct)
+        rc_decrement(dev->icc_struct, "finalize subclass device");
+    if (dev->PageList)
+        rc_decrement(dev->PageList, "finalize subclass device");
+}
