@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2018 Artifex Software, Inc.
+/* Copyright (C) 2001-2019 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -40,7 +40,7 @@ gs_setblendmode(gs_gstate *pgs, gs_blend_mode_t mode)
     if (gs_debug_c('v')) {
         static const char *const bm_names[] = { GS_BLEND_MODE_NAMES };
 
-        dmlprintf1(pgs->memory, "[v](0x%lx)blend_mode = ", (ulong)pgs);
+        dmlprintf1(pgs->memory, "[v](0x%lx)blend_mode = ", (long)pgs);
         if (mode >= 0 && mode < countof(bm_names))
             dmprintf1(pgs->memory, "%s\n", bm_names[mode]);
         else
@@ -155,7 +155,7 @@ check_for_nontrans_pattern(gs_gstate *pgs, unsigned char *comp_name)
  * compositor device.
  */
 static int
-gs_gstate_update_pdf14trans(gs_gstate * pgs, gs_pdf14trans_params_t * pparams)
+gs_gstate_update_pdf14trans2(gs_gstate * pgs, gs_pdf14trans_params_t * pparams, bool retain_on_create)
 {
     gx_device * dev = pgs->device;
     gx_device *pdf14dev = NULL;
@@ -174,6 +174,7 @@ gs_gstate_update_pdf14trans(gs_gstate * pgs, gs_pdf14trans_params_t * pparams)
      */
     if (pdf14dev != dev) {
         gx_set_device_only(pgs, pdf14dev);
+        gx_device_retain(pdf14dev, retain_on_create);
     }
 
     /* If we had a color space change and we are in overprint, then we need to
@@ -185,6 +186,12 @@ gs_gstate_update_pdf14trans(gs_gstate * pgs, gs_pdf14trans_params_t * pparams)
     return code;
 }
 
+static int
+gs_gstate_update_pdf14trans(gs_gstate * pgs, gs_pdf14trans_params_t * pparams)
+{
+    return gs_gstate_update_pdf14trans2(pgs, pparams, true);
+}
+
 void
 gs_trans_group_params_init(gs_transparency_group_params_t *ptgp)
 {
@@ -210,7 +217,7 @@ gs_update_trans_marking_params(gs_gstate * pgs)
 int
 gs_begin_transparency_group(gs_gstate *pgs,
                             const gs_transparency_group_params_t *ptgp,
-                            const gs_rect *pbbox)
+                            const gs_rect *pbbox, pdf14_compositor_operations group_type)
 {
     gs_pdf14trans_params_t params = { 0 };
     const gs_color_space *blend_color_space;
@@ -225,7 +232,7 @@ gs_begin_transparency_group(gs_gstate *pgs,
      * create_compositor.  This will pass the data to the PDF 1.4
      * transparency device.
      */
-    params.pdf14_op = PDF14_BEGIN_TRANS_GROUP;
+    params.pdf14_op = group_type;
     params.Isolated = ptgp->Isolated;
     params.Knockout = ptgp->Knockout;
     params.image_with_SMask = ptgp->image_with_SMask;
@@ -257,9 +264,9 @@ gs_begin_transparency_group(gs_gstate *pgs,
        target device (process color model).  Here we just want
        to set it as a unknown type for clist writing, as we will take care
        of using the parent group color space later during clist reading.
-       Also, if the group was not isolated we MUST use the parent group
-       color space regardless of what the group color space is specified to be
-       */
+     */
+
+
     if (ptgp->ColorSpace == NULL || params.Isolated != true) {
         params.group_color = UNKNOWN;
         params.group_color_numcomps = 0;
@@ -548,6 +555,7 @@ gs_begin_transparency_mask(gs_gstate * pgs,
     int i, code;
     gs_color_space *blend_color_space;
     gsicc_manager_t *icc_manager = pgs->icc_manager;
+    bool deep = device_is_deep(pgs->device);
 
     if (check_for_nontrans_pattern(pgs,
                   (unsigned char *)"gs_pop_transparency_state")) {
@@ -592,7 +600,7 @@ gs_begin_transparency_mask(gs_gstate * pgs,
         if (blend_color_space == NULL)
             return_error(gs_error_VMerror);
         blend_color_space->cmm_icc_profile_data = pgs->icc_manager->default_gray;
-        rc_increment(blend_color_space->cmm_icc_profile_data);
+        gsicc_adjust_profile_rc(blend_color_space->cmm_icc_profile_data, 1, "gs_begin_transparency_mask");
         if_debug9m('v', pgs->memory, "[v](0x%lx)gs_begin_transparency_mask [%g %g %g %g]\n\
           subtype = %d  Background_components = %d, Matte_components = %d, %s\n",
                   (ulong)pgs, pbbox->p.x, pbbox->p.y, pbbox->q.x, pbbox->q.y,
@@ -601,12 +609,29 @@ gs_begin_transparency_mask(gs_gstate * pgs,
                   (ptmp->TransferFunction == mask_transfer_identity ? "no TR" :
                    "has TR"));
         /* Sample the transfer function */
-        for (i = 0; i < MASK_TRANSFER_FUNCTION_SIZE; i++) {
-            float in = (float)(i * (1.0 / (MASK_TRANSFER_FUNCTION_SIZE - 1)));
+        /* For non-deep cases, we sample at 00,01,02..fe,ff.
+         * For deep cases, we sample from 0000,0100,0200..fe00,ff00 and a final one at ffff.
+         * This enables us to interpolate easily.
+         */
+        if (deep) {
+            uint16_t *trans16 = (uint16_t *)params.transfer_fn;
             float out;
+            for (i = 0; i < MASK_TRANSFER_FUNCTION_SIZE; i++) {
+                float in = (float)(i * (1.0 / MASK_TRANSFER_FUNCTION_SIZE));
 
-            ptmp->TransferFunction(in, &out, ptmp->TransferFunction_data);
-            params.transfer_fn[i] = (byte)floor((double)(out * 255 + 0.5));
+                ptmp->TransferFunction(in, &out, ptmp->TransferFunction_data);
+                trans16[i] = (uint16_t)floor((double)(out * 65535 + 0.5));
+            }
+            ptmp->TransferFunction(1.0, &out, ptmp->TransferFunction_data);
+            trans16[MASK_TRANSFER_FUNCTION_SIZE] = (uint16_t)floor((double)(out * 65535 + 0.5));
+        } else {
+            for (i = 0; i < MASK_TRANSFER_FUNCTION_SIZE; i++) {
+                float in = (float)(i * (1.0 / (MASK_TRANSFER_FUNCTION_SIZE - 1)));
+                float out;
+
+                ptmp->TransferFunction(in, &out, ptmp->TransferFunction_data);
+                params.transfer_fn[i] = (byte)floor((double)(out * 255 + 0.5));
+            }
         }
         /* Note:  This function is called during the c-list writer side. */
         if ( blend_color_space->cmm_icc_profile_data != NULL ) {
@@ -628,7 +653,7 @@ gs_begin_transparency_mask(gs_gstate * pgs,
         /* Explicitly decrement the profile data since blend_color_space may not
          * be an ICC color space object.
          */
-        rc_decrement(blend_color_space->cmm_icc_profile_data, "gs_begin_transparency_mask");
+        gsicc_adjust_profile_rc(blend_color_space->cmm_icc_profile_data, -1, "gs_begin_transparency_mask");
         rc_decrement_only_cs(blend_color_space, "gs_begin_transparency_mask");
     }
     return gs_gstate_update_pdf14trans(pgs, &params);
@@ -763,7 +788,7 @@ get_num_pdf14_spot_colors(gs_gstate * pgs)
 }
 
 int
-gs_push_pdf14trans_device(gs_gstate * pgs, bool is_pattern)
+gs_push_pdf14trans_device(gs_gstate * pgs, bool is_pattern, bool retain)
 {
     gs_pdf14trans_params_t params = { 0 };
     cmm_profile_t *icc_profile;
@@ -794,7 +819,7 @@ gs_push_pdf14trans_device(gs_gstate * pgs, bool is_pattern)
         params.iccprofile = pgs->icc_manager->default_rgb;
     }
     /* Note: Other parameters not used */
-    return gs_gstate_update_pdf14trans(pgs, &params);
+    return gs_gstate_update_pdf14trans2(pgs, &params, retain);
 }
 
 int
